@@ -1,10 +1,10 @@
-# $Id: PropertyList.pm,v 1.7 2002/11/12 22:39:58 comdog Exp $
+# $Id: PropertyList.pm,v 1.8 2002/12/18 10:35:56 comdog Exp $
 package Mac::PropertyList;
 use strict;
 
 use vars qw($ERROR $XML_head $XML_foot $VERSION);
 
-$VERSION = 0.07;
+$VERSION = 0.08;
 
 =head1 NAME
 
@@ -13,13 +13,13 @@ Mac::PropertyList - work with Mac plists at a low level
 =head1 SYNOPSIS
 
 	use Mac::PropertyList;
-	
+
 	my $data  = parse_plist( $text );
 
 	my $text  = plist_as_string( $data );
-	
+
 	my $plist = create_from_hash( \%hash );
-	
+
 =head1 DESCRIPTION
 
 This module is a low-level interface to the Mac OS X
@@ -35,7 +35,7 @@ that's your business. :)
 
 =head2 The Property List format
 
-The MacOS X Property List format is simple XML.  You 
+The MacOS X Property List format is simple XML.  You
 can read the DTD to get the details.
 
 	http://www.apple.com/DTDs/PropertyList-1.0.dtd
@@ -45,7 +45,7 @@ structure to list keys and values so that values
 are only associated with their keys by their position
 in the file rather than by the structure of the DTD.
 This problem is the major design hinderance in this
-module.  A smart XML format would have made things 
+module.  A smart XML format would have made things
 much easier.
 
 =head2 The Mac::PropertyList data structure
@@ -70,7 +70,7 @@ The hash for a string object looks like
 The structure for the date, data, integer, float, and
 real look the same.
 
-The plist objects true and false are wierd since in 
+The plist objects true and false are wierd since in
 the XML they are empty elements.  Mac::PropertyList
 makes them look not-empty.  This may seem wierd, but
 it saved hours of work in the implementation.
@@ -98,7 +98,7 @@ of the keys are in turn plist objects again.
 	{
 	type  => 'dict',
 	value => {
-		"Bar" => { type => string,  value => 'Foo' } 
+		"Bar" => { type => string,  value => 'Foo' }
 		}
 	}
 
@@ -111,7 +111,7 @@ a simple interface to a particular plist file.
 Run a small script against your favorite plist file
 then dump the results with Data::Dumper.  That's
 what the real deal looks like.
-		
+
 =cut
 
 my $Debug = $ENV{PLIST_DEBUG};
@@ -129,16 +129,16 @@ $XML_foot =<<"XML";
 XML
 
 my %Readers = (
-	"<dict>"    => \&read_dict,
-	"<string>"  => \&read_string,
-	"<date>"    => \&read_date,
-	"<real>"    => \&read_real,
-	"<integer>" => \&read_integer,
-	"<string>"  => \&read_string,
-	"<array>"   => \&read_array,
-	"<data>"    => \&read_data,
-	"<true>"    => \&read_true,
-	"<false>"   => \&read_false,
+	"dict"    => \&read_dict,
+	"string"  => \&read_string,
+	"date"    => \&read_date,
+	"real"    => \&read_real,
+	"integer" => \&read_integer,
+	"string"  => \&read_string,
+	"array"   => \&read_array,
+	"data"    => \&read_data,
+	"true"    => \&read_true,
+	"false"   => \&read_false,
 	);
 
 my %Writers = (
@@ -169,54 +169,29 @@ data structure.
 sub _strip_leading_space
 	{
 	my $ref = shift;
-	
-	$$ref =~ s/^\s*//;	
+
+	$$ref =~ s/^\s*//;
 	}
-	
+
 sub _get_next_tag
 	{
 	my $ref = shift;
 	_strip_leading_space($ref);
-	
+
 	my( $tag ) = $$ref =~ m/^(<.*?>)/g;
-	
+
 	return $tag;
 	}
-	
+
 sub parse_plist
 	{
 	my $text = shift;
-	
-	$text =~ s|<true/>|<true>true</true>|g;
-	$text =~ s|<false/>|<false>false</false>|g;
 
 	# we can handle either 0.9 or 1.0
 	$text =~ s|^<\?xml.*?>\s*<!DOC.*>\s*<plist.*?>\s*||;
 	$text =~ s|\s*</plist>\s*$||;
-	
-	my $plist = do {
-		if( $text =~ s/^\s*<(array|dict)>\s*// )
-			{
-			my $type = $1;
-			$text =~ s|\s*</$type>\s*$||;
-			$Readers{"<$type>"}->( \$text );
-			}
-		else
-			{
-			my @object = extract_tagged( $text, 
-				undef, undef, undef, $Options );
-			
-			unless( exists $Readers{$object[3]} )
-				{
-				$ERROR = "Not a plist object [$object[3]]!";
-				return;
-				}
-				
-			print STDERR "Found $object[3]\n" if $Debug > 1;
-			
-			$Readers{$object[3]}->($object[4]);
-			}
-		};		
+
+	my $plist = read_next( \$text );
 
 	return $plist;
 	}
@@ -235,28 +210,28 @@ Returns a string representing the hash in the plist format.
 sub create_from_hash
 	{
 	my $hash  = shift;
-	
+
 	return unless UNIVERSAL::isa( $hash, 'HASH' );
-	
+
 	my $string = "$XML_head<dict>\n";
-	
+
 	foreach my $key ( keys %$hash )
 		{
 		my( $type, $value ) = ( 'string', $hash->{$key} );
-		
+
 		next if ref $value;
-		
+
 		my $bit  = _string( 'key', $key ) . "\n";
 		   $bit .= $Writers{$type}->( $value ) . "\n";
-		
+
 		$bit =~ s/^/\t/gm;
-		
+
 		$string .= $bit;
 		}
-	
+
 	$string .= "</dict>\n$XML_foot";
-	
-	return $string;		
+
+	return $string;
 	}
 
 sub _hash { { type => $_[0], value => $_[1] } }
@@ -265,95 +240,96 @@ sub _read_tag
 	{
 	my $tag = shift;
 	my $ref = shift;
-	
+
 	$$ref = s|\s*<$tag>(.*?)</$tag>\s*||g;
-	
+
 	return $Readers{"<$tag>"}->($1);
 	}
-	
+
 sub read_string  { _hash( 'string',  $_[0] ) }
 sub read_integer { _hash( 'integer', $_[0] ) }
 sub read_date    { _hash( 'date',    $_[0] ) }
 sub read_real    { _hash( 'real',    $_[0] ) }
-sub read_true    { _hash( 'true',    $_[0] ) }
-sub read_false   { _hash( 'false',   $_[0] ) }
+sub read_true    { _hash( 'true',  'true'  ) }
+sub read_false   { _hash( 'false', 'false' ) }
 
 sub read_next
 	{
 	my $ref = shift;
-	
+
 	my $value = do {
 		my $tag;
 		my $value;
-		if( $$ref =~ m[^\s*<(string|date|real|integer|data|true|false)>\s*(.*?)\s*</\1>]s )
+		if( $$ref =~ s[^\s* < (string|date|real|integer|data) > 
+			\s*(.*?)\s* </\1> ][]sx )
 			{
-			$tag   = $1;
-			$value = $2;
-			print STDERR "read_next: Found value type $1\n" if $Debug > 1;
-			$$ref =~ s|\s*<$tag>\s*(.*?)\s*</$tag>\s*||s;
-			_print_next_bit( $ref ) if $Debug > 1;
-			$Readers{"<$tag>"}->($value);
+			print STDERR "read_next: Found value type [$1] with value [$2]\n" 
+				if $Debug > 1;
+			$Readers{$1}->( $2 );
 			}
-		elsif( $$ref =~ m[^\s*<(dict|array)>(.*?)</\1>]s and $tag = $1 and $value = $2 and $value !~ m/<$tag>/ )
+		elsif( $$ref =~ s[^\s* < (dict|array) > ][]x )
 			{
-			print STDERR "dict: Found value type $1 without nested $1\n" if $Debug > 1;
-			$$ref =~ s|\s*<$tag>\s*(.*?)\s*</$tag>\s*||s;
-			_print_next_bit( $ref ) if $Debug > 1;
-			$Readers{"<$tag>"}->(\$value);
+			print STDERR "\$1 is [$1]\n" if $Debug > 1;
+			$Readers{$1}->( $ref );
+			}
+		# these next two are some wierd cases i found in the iPhoto Preferences
+		elsif( $$ref =~ s[^\s* < dict / > ][]x )
+			{
+			return _hash( 'dict', {} );
+			}
+		elsif( $$ref =~ s[^\s* < array / > ][]x )
+			{
+			return _hash( 'array', [] );
+			}
+		elsif( $$ref =~ s[^\s* < (true|false) /> ][]x )
+			{
+			print STDERR "\$1 is [$1]\n" if $Debug > 1;
+			$Readers{$1}->();
 			}
 		};
 
 	}
-	
+
 sub read_dict
 	{
 	my $ref = shift;
 	print STDERR "Processing dict\n" if $Debug > 1;
-	
+
 	my %hash;
-	
-	while( $$ref =~ s|^\s*<key>(.*?)</key>\s*||s )
+
+	while( not $$ref =~ s|^\s*</dict>|| )
 		{
+		$$ref =~ s[^\s*<key>(.*?)</key>][]s;
+		print STDERR "read_dict: key is [$1]\n" if $Debug > 1;
+		die "Could not read key!" unless $1;
 		my $key = $1;
-						
-		$hash{$key} = read_next( $ref );
+		$hash{ $key } = read_next( $ref );
 		}
-		
+
 	return _hash( 'dict', \%hash );
 	}
 
-sub _print_next_bit
-	{
-	my $ref = shift;
-	my $sub = (caller(1))[3];
-	$sub =~ s/.*:://;
-	
-	print STDERR "$sub: Next bit is [" . substr( $$ref, 0, 10 ) . "]\n";
-	}
-	
 sub read_array
 	{
 	my $ref = shift;
-	
 	print STDERR "Processing array\n" if $Debug > 1;
-	
+
 	my @array = ();
 
-	while( $$ref )
+	while( not $$ref =~ s|^\s*</array>|| )
 		{
-		_print_next_bit( $ref ) if $Debug > 1;
-		push @array, read_next( $ref )
+		push @array, read_next( $ref );
 		}
-		
+
 	return _hash( 'array', \@array );
 	}
 
 sub read_data
 	{
 	my $string = shift;
-	
+
 	require MIME::Base64;
-	
+
 	$string = MIME::Base64::decode_base64($string);
 
 	return _hash( 'data', $string );
@@ -366,14 +342,14 @@ sub plist_as_string
 	my( $type, $value ) = @{ $hash }{ qw( type value ) };
 
 	my $string = $XML_head;
-	
+
 	$string .= $Writers{$type}->($value) . "\n";
-	
+
 	$string .= $XML_foot;
-	
+
 	return $string;
 	}
-	
+
 sub _string { "<$_[0]>$_[1]</$_[0]>" }
 
 sub write_string  { _string( 'string',  $_[0] ) }
@@ -386,9 +362,9 @@ sub write_false   { "<false/>" }
 sub write_data($)
 	{
 	my $string = shift;
-	
+
 	require MIME::Base64;
-	
+
 	$string = MIME::Base64::encode_base64($string);
 
 	return _string( 'data', $string );
@@ -397,46 +373,46 @@ sub write_data($)
 sub write_array
 	{
 	my $array = shift;
-	
+
 	my $string = "<array>\n";
-	
+
 	foreach my $element ( @$array )
 		{
 		my( $type, $value ) = @{ $element }{ qw( type value ) };
-		
+
 		my $bit = $Writers{$type}->( $value );
-		
+
 		$bit =~ s/^/\t/gm;
-		
+
 		$string .= $bit . "\n";
 		}
-	
+
 	$string .= "</array>";
-	
-	return $string;		
+
+	return $string;
 	}
 
 sub write_dict
 	{
 	my $dict  = shift;
-	
+
 	my $string = "<dict>\n";
-	
+
 	foreach my $key ( keys %$dict )
 		{
 		my( $type, $value ) = @{ $dict->{$key} }{ qw( type value ) };
-		
+
 		my $bit  = _string( 'key', $key ) . "\n";
 		   $bit .= $Writers{$type}->( $value ) . "\n";
-		
+
 		$bit =~ s/^/\t/gm;
-		
+
 		$string .= $bit;
 		}
-	
+
 	$string .= "</dict>";
-	
-	return $string;		
+
+	return $string;
 	}
 
 =back
@@ -447,30 +423,19 @@ This source is part of a SourceForge project which always has the
 latest sources in CVS, as well as all of the previous releases.
 
 	https://sourceforge.net/projects/brian-d-foy/
-	
+
 If, for some reason, I disappear from the world, one of the other
 members of the project can shepherd this module appropriately.
 
+=head1 CREDITS
+
+Thanks to Chris Nandor for general Mac kung fu and Chad Walker
+for help figuring out the recursion for nested structures.
+
 =head1 TO DO
 
 * actually test the write_* stuff
 
-=head1 BUGS
-
-* i've taken some shortcuts with the parsing, since balanced text parsing 
-can be really slow.  at the moment this module can't handle more than
-one level of nest dicts or arrays.  this breaks on some application's
-files:
-
-	com.apple.DiskCopy.plist
-	com.apple.dock.plist
-	com.apple.Preview.plist
-	com.apple.TextEdit.plist
-	org.aegidian.yaxjournal.plist
-	FruitMenu.prefPane/Contents/Resources/Library.plist
-	FruitMenu.prefPane/Contents/Resources/Presets.plist
-
-
 =head1 AUTHOR
 
 brian d foy, E<lt>bdfoy@cpan.orgE<gt>
